"""Gremlin Server exceptions."""

__all__ = ("RequestError", "GremlinServerError")


class StatusException(IOError):

    def __init__(self, value, result):
        """Handle all exceptions returned from the Gremlin Server as per:
        """
        self.value = value
        self.response = {
            498: ("MALFORMED_REQUEST",
                  ("The request message was not properly formatted which " +
                   "means it could not be parsed at all or the 'op' code " +
                   "was not recognized such that Gremlin Server could " +
                   "properly route it for processing. Check the message " +
                   "format and retry the request")),
            499: ("INVALID_REQUEST_ARGUMENTS",
                  ("The request message was parseable, but the arguments " +
                   "supplied in the message were in conflict or incomplete. " +
                   "Check the message format and retry the request.")),
            500: ("SERVER_ERROR",
                  ("A general server error occurred that prevented the " +
                   "request from being processed.")),
            596: ("TRAVERSAL_EVALUATION",
                  ("The remote " +
                   "{@link org.apache.tinkerpop.gremlin.process.Traversal} " +
                   "submitted for processing evaluated in on the server " +
                   "with errors and could not be processed")),
            597: ("SCRIPT_EVALUATION",
                  ("The script submitted for processing evaluated in the " +
                   "{@code ScriptEngine} with errors and could not be  " +
                   "processed.Check the script submitted for syntax errors " +
                   "or other problems and then resubmit.")),
            598: ("TIMEOUT",
                  ("The server exceeded one of the timeout settings for the " +
                   "request and could therefore only partially respond or " +
                   " not respond at all.")),
            599: ("SERIALIZATION",
                  ("The server was not capable of serializing an object " +
                   "that was returned from the script supplied on the " +
                   "requst. Either transform the object into something " +
                   "Gremlin Server can process within the script or install " +
                   "mapper serialization classes to Gremlin Server."))
        }
        if result:
            result = "\n\n{}".format(result)
        self.message = 'Code [{}]: {}. {}.{}'.format(
            self.value,
            self.response[self.value][0],
            self.response[self.value][1],
            result)
        super().__init__(self.message)


class RequestError(StatusException):
    pass


class GremlinServerError(StatusException):
    pass



"""
"""

import asyncio
import unittest
import uuid

from aiogremlin import (submit, GremlinConnector, GremlinClient,
                        GremlinClientSession)


class SubmitTest(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)

    def tearDown(self):
        self.loop.close()

    def test_submit(self):

        @asyncio.coroutine
        def go():
            resp = yield from submit("4 + 4", bindings={"x": 4},
                                     loop=self.loop)
            results = yield from resp.get()
            return results

        results = self.loop.run_until_complete(go())
        self.assertEqual(results[0].data[0], 8)


class GremlinClientTest(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.gc = GremlinClient(url="ws://localhost:8182/", loop=self.loop)

    def tearDown(self):
        self.loop.run_until_complete(self.gc.close())
        self.loop.close()

    def test_connection(self):

        @asyncio.coroutine
        def go():
            ws = yield from self.gc._connector.ws_connect(self.gc.url)
            self.assertFalse(ws.closed)
            yield from ws.close()

        self.loop.run_until_complete(go())

    def test_execute(self):

        @asyncio.coroutine
        def go():
            resp = yield from self.gc.execute("x + x", bindings={"x": 4})
            return resp

        results = self.loop.run_until_complete(go())
        self.assertEqual(results[0].data[0], 8)

    def test_sub_waitfor(self):
        sub1 = self.gc.execute("x + x", bindings={"x": 1})
        sub2 = self.gc.execute("x + x", bindings={"x": 2})
        sub3 = self.gc.execute("x + x", bindings={"x": 4})
        coro = asyncio.gather(*[asyncio.async(sub1, loop=self.loop),
                              asyncio.async(sub2, loop=self.loop),
                              asyncio.async(sub3, loop=self.loop)],
                              loop=self.loop)
        # Here I am looking for resource warnings.
        results = self.loop.run_until_complete(coro)
        self.assertIsNotNone(results)

    def test_resp_stream(self):
        @asyncio.coroutine
        def stream_coro():
            results = []
            resp = yield from self.gc.submit("x + x", bindings={"x": 4})
            while True:
                f = yield from resp.stream.read()
                if f is None:
                    break
                results.append(f)
            self.assertEqual(results[0].data[0], 8)
        self.loop.run_until_complete(stream_coro())

    def test_execute_error(self):
        execute = self.gc.execute("x + x g.asdfas", bindings={"x": 4})
        try:
            self.loop.run_until_complete(execute)
            error = False
        except:
            error = True
        self.assertTrue(error)


class GremlinClientSessionTest(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.gc = GremlinClientSession(url="ws://localhost:8182/",
                                       loop=self.loop)
        self.script1 = """graph = TinkerFactory.createModern()
                          g = graph.traversal(standard())"""

        self.script2 = "g.V().has('name','marko').out('knows').values('name')"

    def tearDown(self):
        self.loop.run_until_complete(self.gc.close())
        self.loop.close()

    def test_session(self):

        @asyncio.coroutine
        def go():
            yield from self.gc.execute(self.script1)
            results = yield from self.gc.execute(self.script2)
            return results

        results = self.loop.run_until_complete(go())
        self.assertTrue(len(results[0].data), 2)

    def test_session_reset(self):

        @asyncio.coroutine
        def go():
            yield from self.gc.execute(self.script1)
            self.gc.reset_session()
            results = yield from self.gc.execute(self.script2)
            return results

        results = self.loop.run_until_complete(go())
        self.assertIsNone(results[0].data)

    def test_session_manual_reset(self):

        @asyncio.coroutine
        def go():
            yield from self.gc.execute(self.script1)
            new_sess = str(uuid.uuid4())
            sess = self.gc.reset_session(session=new_sess)
            self.assertEqual(sess, new_sess)
            self.assertEqual(self.gc.session, new_sess)
            results = yield from self.gc.execute(self.script2)
            return results

        results = self.loop.run_until_complete(go())
        self.assertIsNone(results[0].data)

    def test_session_set(self):

        @asyncio.coroutine
        def go():
            yield from self.gc.execute(self.script1)
            new_sess = str(uuid.uuid4())
            self.gc.session = new_sess
            self.assertEqual(self.gc.session, new_sess)
            results = yield from self.gc.execute(self.script2)
            return results

        results = self.loop.run_until_complete(go())
        self.assertIsNone(results[0].data)

    def test_resp_session(self):

        @asyncio.coroutine
        def go():
            session = str(uuid.uuid4())
            self.gc.session = session
            resp = yield from self.gc.submit("x + x", bindings={"x": 4})
            while True:
                f = yield from resp.stream.read()
                if f is None:
                    break
            self.assertEqual(resp.session, session)

        self.loop.run_until_complete(go())


class ContextMngrTest(unittest.TestCase):

    def setUp(self):
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(None)
        self.connector = GremlinConnector(loop=self.loop)

    def tearDown(self):
        self.loop.run_until_complete(self.connector.close())
        self.loop.close()

    # def test_connection_manager(self):
    #     results = []
    #
    #     @asyncio.coroutine
    #     def go():
    #         with (yield from self.connector) as conn:
    #             client = SimpleGremlinClient(conn, loop=self.loop)
    #             resp = yield from client.submit("1 + 1")
    #             while True:
    #                 mssg = yield from resp.stream.read()
    #                 if mssg is None:
    #                     break
    #                 results.append(mssg)
    #     self.loop.run_until_complete(go())


if __name__ == "__main__":
    unittest.main()



"""Websocket connection factory and manager."""

import asyncio

from aiowebsocketclient import WebSocketConnector

from aiogremlin.response import GremlinClientWebSocketResponse

__all__ = ("GremlinConnector",)


class GremlinConnector(WebSocketConnector):
    """Create and manage reusable websocket connections. Out of the box
    support for multiple enpoints (databases).

    :param float conn_timeout: timeout for establishing connection (optional).
        Values ``0`` or ``None`` mean no timeout
    :param bool force_close: close websockets after release
    :param int limit: limit for total open websocket connections
    :param aiohttp.client.ClientSession client_session: Underlying HTTP
        session used to establish websocket connections
    :param loop: `event loop` If param is ``None``, `asyncio.get_event_loop`
        is used for getting default event loop (optional)
    :param ws_response_class: WebSocketResponse class implementation.
        ``ClientWebSocketResponse`` by default
    :param bool verbose: Set log level to info. False by default
    """
    def __init__(self, *, conn_timeout=None, force_close=False, limit=1024,
                 client_session=None, loop=None, verbose=False):

        super().__init__(conn_timeout=conn_timeout, force_close=force_close,
                         limit=limit, client_session=client_session, loop=loop,
                         ws_response_class=GremlinClientWebSocketResponse)



"""Client for the Tinkerpop 3 Gremlin Server."""

import asyncio
import uuid

import aiohttp

from aiogremlin.response import GremlinClientWebSocketResponse
from aiogremlin.exceptions import RequestError
from aiogremlin.connector import GremlinConnector
from aiogremlin.subprotocol import gremlin_response_parser, GremlinWriter

__all__ = ("submit", "GremlinClient", "GremlinClientSession",
           "GremlinResponse", "GremlinResponseStream")


class GremlinClient:
    """Main interface for interacting with the Gremlin Server.

    :param str url: url for Gremlin Server (optional). 'ws://localhost:8182/'
        by default
    :param loop: :ref:`event loop<asyncio-event-loop>` If param is ``None``,
        `asyncio.get_event_loop` is used for getting default event loop
        (optional)
    :param str lang: Language of scripts submitted to the server.
        "gremlin-groovy" by default
    :param str op: Gremlin Server op argument. "eval" by default.
    :param str processor: Gremlin Server processor argument. "" by default.
    :param float timeout: timeout for establishing connection (optional).
        Values ``0`` or ``None`` mean no timeout
    :param connector: A class that implements the method ``ws_connect``.
        Usually an instance of ``aiogremlin.connector.GremlinConnector``
    """

    def __init__(self, *, url='ws://localhost:8182/', loop=None,
                 lang="gremlin-groovy", op="eval", processor="",
                 timeout=None, ws_connector=None):
        """
        """
        self._lang = lang
        self._op = op
        self._processor = processor
        self._loop = loop or asyncio.get_event_loop()
        self._closed = False
        self._session = None
        self._url = url
        self._timeout = timeout
        if ws_connector is None:
            ws_connector = GremlinConnector(loop=self._loop)
        self._connector = ws_connector

    @property
    def loop(self):
        """Readonly property that returns event loop used by client"""
        return self._loop

    @property
    def op(self):
        """Readonly property that returns op argument for Gremlin Server"""
        return self._op

    @property
    def processor(self):
        """Readonly property. The processor argument for Gremlin
        Server"""
        return self._processor

    @property
    def lang(self):
        """Readonly property. The language used for Gremlin scripts"""
        return self._lang

    @property
    def url(self):
        """Getter/setter for database url used by the client"""
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

    @property
    def closed(self):
        """Readonly property. Return True if client has been closed"""
        return self._closed or self._connector is None

    @asyncio.coroutine
    def close(self):
        """Close client. If client has not been detached from underlying
        ws_connector, this coroutinemethod closes the latter as well."""
        if self._closed:
            return
        self._closed = True
        try:
            yield from self._connector.close()
        finally:
            self._connector = None

    def detach(self):
        """Detach client from ws_connector. Client status is now closed"""
        self._connector = None

    @asyncio.coroutine
    def submit(self, gremlin, *, bindings=None, lang=None,
               op=None, processor=None, binary=True, session=None,
               timeout=None):
        """
        """
        lang = lang or self.lang
        op = op or self.op
        processor = processor or self.processor
        if session is None:
            session = self._session
        if timeout is None:
            timeout = self._timeout

        ws = yield from self._connector.ws_connect(
            self.url, timeout=timeout)

        writer = GremlinWriter(ws)

        ws = writer.write(gremlin, bindings=bindings, lang=lang, op=op,
                          processor=processor, binary=binary,
                          session=session)

        return GremlinResponse(ws, session=session, loop=self._loop)

    @asyncio.coroutine
    def execute(self, gremlin, *, bindings=None, lang=None, session=None,
                op=None, processor=None, binary=True, timeout=None):
        """
        """
        lang = lang or self.lang
        op = op or self.op
        processor = processor or self.processor
        resp = yield from self.submit(gremlin, bindings=bindings, lang=lang,
                                      op=op, processor=processor,
                                      binary=binary, session=session,
                                      timeout=timeout)

        return (yield from resp.get())


class GremlinClientSession(GremlinClient):
    """Interface for interacting with the Gremlin Server using sessions.

    :param str url: url for Gremlin Server (optional). 'ws://localhost:8182/'
        by default
    :param loop: :ref:`event loop<asyncio-event-loop>` If param is ``None``,
        `asyncio.get_event_loop` is used for getting default event loop
        (optional)
    :param str lang: Language of scripts submitted to the server.
        "gremlin-groovy" by default
    :param str op: Gremlin Server op argument. "eval" by default.
    :param str processor: Gremlin Server processor argument. "" by default.
    :param float timeout: timeout for establishing connection (optional).
        Values ``0`` or ``None`` mean no timeout
    """

    def __init__(self, *, url='ws://localhost:8182/', loop=None,
                 lang="gremlin-groovy", op="eval", processor="session",
                 session=None, timeout=None,
                 ws_connector=None):
        super().__init__(url=url, lang=lang, op=op, processor=processor,
                         loop=loop, timeout=timeout, ws_connector=ws_connector)

        if session is None:
            session = str(uuid.uuid4())
        self._session = session

    @property
    def session(self):
        return self._session

    @session.setter
    def session(self, value):
        self._session = value

    def reset_session(self, session=None):
        if session is None:
            session = str(uuid.uuid4())
        self._session = session
        return self._session


class GremlinResponse:
    """Main interface for reading Gremlin Server responses. Typically returned
    by ``GremlinClient.submit``, not created by user.

    :param ``aiogremlin.response.GremlinClientWebSocketResponse`` ws: Websocket
        connection.
    :param str session: Session id (optional). Typically a uuid
    :param loop: :ref:`event loop<asyncio-event-loop>` If param is ``None``,
        `asyncio.get_event_loop` is used for getting default event loop
        (optional)
    """
    def __init__(self, ws, *, session=None, loop=None):
        self._loop = loop or asyncio.get_event_loop()
        self._session = session
        self._stream = GremlinResponseStream(ws, loop=self._loop)

    @property
    def stream(self):
        return self._stream

    @property
    def session(self):
        return self._session

    @asyncio.coroutine
    def get(self):
        return (yield from self._run())

    @asyncio.coroutine
    def _run(self):
        """
        """
        results = []
        while True:
            message = yield from self._stream.read()
            if message is None:
                break
            results.append(message)
        return results


class GremlinResponseStream:
    """
    Encapsulate and read Gremlin Server responses. Typically instantiated by
    GremlinResponse constructor, not by user.

    :param ``aiogremlin.response.GremlinClientWebSocketResponse`` ws: Websocket
        connection.
    :param loop: :ref:`event loop<asyncio-event-loop>` If param is ``None``,
        `asyncio.get_event_loop` is used for getting default event loop
        (optional)
    """
    def __init__(self, ws, loop=None):
        self._ws = ws
        self._loop = loop or asyncio.get_event_loop()
        data_stream = aiohttp.DataQueue(loop=self._loop)
        self._stream = self._ws.parser.set_parser(gremlin_response_parser,
                                                  output=data_stream)

    @asyncio.coroutine
    def read(self):
        """Read a message from the stream"""
        if self._stream.at_eof():
            yield from self._ws.release()
            message = None
        else:
            asyncio.async(self._ws.receive(), loop=self._loop)
            try:
                message = yield from self._stream.read()
            except RequestError:
                yield from self._ws.release()
                raise
        return message


@asyncio.coroutine
def submit(gremlin, *,
           url='ws://localhost:8182/',
           bindings=None,
           lang="gremlin-groovy",
           op="eval",
           processor="",
           timeout=None,
           session=None,
           loop=None):
    """Submit a script to the Gremlin Server.

    :param str url: url for Gremlin Server (optional). 'ws://localhost:8182/'
        by default
    :param str lang: Language of scripts submitted to the server.
        "gremlin-groovy" by default
    :param str op: Gremlin Server op argument. "eval" by default.
    :param str processor: Gremlin Server processor argument. "" by default.
    :param float timeout: timeout for establishing connection (optional).
        Values ``0`` or ``None`` mean no timeout
    :param str session: Session id (optional). Typically a uuid
    :param loop: :ref:`event loop<asyncio-event-loop>` If param is ``None``,
        `asyncio.get_event_loop` is used for getting default event loop
        (optional)
    """

    if loop is None:
        loop = asyncio.get_event_loop()

    connector = aiohttp.TCPConnector(force_close=True, loop=loop)

    client_session = aiohttp.ClientSession(
        connector=connector, loop=loop,
        ws_response_class=GremlinClientWebSocketResponse)

    gremlin_client = GremlinClient(url=url, loop=loop,
                                   ws_connector=client_session)

    try:
        resp = yield from gremlin_client.submit(
            gremlin, bindings=bindings, lang=lang, op=op, processor=processor,
            session=session, timeout=timeout)

        return resp

    finally:
        gremlin_client.detach()
        client_session.detach()
